// DXF to POV converter (version 0.81) : DXF input module
// Coded by Richard Dorman (March-April 1994)

// Notes : Surface normals are calculated anti-clockwise.
//       : Quadrelaterals are assunmed to be planar.

#include "DXF-POV.h"

#include "conio.h"
#include "graphics.h"

#define HEADER   "HEADER"
#define TABLES   "TABLES"
#define BLOCKS   "BLOCKS"
#define ENTITIES "ENTITIES"
#define VIEW     "VIEW"
#define ENDSEC   "ENDSEC"
#define ENDTAB   "ENDTAB"
#define ENDBLK   "ENDBLK"
#define SEQEND   "SEQEND"
#define CIRCLE   "CIRCLE"
#define FACE3D   "3DFACE"
#define POLYLINE "POLYLINE"
#define INSERT   "INSERT"
#define	FILEEND  "EOF"

#define CAMERA_OPTION   'c'
#define CONVERT_OPTION  'l'
#define GRAPHICS_OPTION 'g'
#define SMOOTH_OPTION   's'

#define ACTIVE "*ACTIVE"

#define READ_COORDINATE(X,C1,C2,C3) \
  case C1:   Assign(X.x,Group); break; \
  case C2:   Assign(X.y,Group); break; \
  case C3:   Assign(X.z,Group); break

#define READ_ENTITY_OPTIONAL(X)        \
  case 8:   Assign(X.Layer,Group);  break; \
  case 5:   Assign(X.Handle,Group); break; \
  case 62:  Assign(X.Colour,Group); break

#define READ_ENTITY_NORMAL(X)  \
  case 210: Assign(X.x,Group); break;\
  case 220: Assign(X.y,Group); break;\
  case 230: Assign(X.z,Group); break

ifstream In;
ofstream Out;

char Inputfile[256],Outputfile[256];

int ConvertLayerNames = TRUE;          // Command line option : Layers
int CameraOn          = TRUE;          // Command line option : Camera
int SmoothTriangles   = TRUE;          // Command line option : Smooth triangles
int GraphicsOn        = FALSE;         // Command line option : Graphics


Group_Type    Group;                   // A group : Code and value

ViewPort_Type View;                    // A view
Vertex_Type   Vertex;                  // A vertex
Circle_Type   Circle;                  // A circle/cylinder
Face3D_Type   Face3D;                  // A 3D face
PolyLine_Type PolyLine;                // A polygon/polyface mesh
Block_Type    Block;                   // A block
Insert_Type   Insert;                  // An insert

point         Mesh[MAXMESH],           // Array of vertices
	      Normals[MAXMESH];        // Array of vertice normals
int	      Faces[MAXMESH][4],Fc[4], // Array of faces (3 or 4 vertices each)
	      VCount[MAXMESH];         // Array of vertice counts


// This is still in development so I have included it direclty.
#include "graphics.cpp"

/* ------------------------------------------------------------------------ */


// The exclude lists contains names of blocks and inserts that must be
// ignored.

#define ExcludeListSize 5
char *ExcludeList[ExcludeListSize] =
  {"AME_SOL","AME_NIL","AME_FRZ","AVE_RENDER","AME_JNK"};

// Checks if name is in the exclude list. Excluded names are also those
// that begin with *.

int InExcludeList(char *Name) {
  int i;

  if (Name[0] == '*') return 1;
  for(i=0;i<ExcludeListSize;i++)
    if (strcmp(Name,ExcludeList[i]) == 0) return 1;
  return 0;
}


/* ------------------------------------------------------------------------ */


void HeaderSection() {

  // Ignore everything

  while (strcmp(Group.value,ENDSEC) != 0)
    In >> Group;
}


/* ------------------------------------------------------------------------ */


void TablesSection() {

  // Find views table.

  while (strcmp(Group.value,VIEW) != 0 && strcmp(Group.value,ENDSEC) != 0)
    In >> Group;

    In >> Group; 	      		  // 70 <Max table entries>
    In >> Group;                          // 0 VIEW

  // Read in views. Each view creates a camera.

  while (strcmp(Group.value,VIEW) == 0) {

    In >> Group;
    while (Group.code != 0) {
      switch (Group.code) {
	case 2:  Assign(View.Name,Group); break;
	case 40: Assign(View.Shift.x,Group); break;
	case 41: Assign(View.Shift.y,Group); break;
	case 42: Assign(View.Lens,Group); break;
	case 50: Assign(View.TwistAngle,Group); break;
	READ_COORDINATE(View.Location,11,21,31);
	READ_COORDINATE(View.LookAt,12,22,32);
      }
      In >> Group;
    }

    WriteCamera(View);
  }

  // Read rest

  In >> Group;         			  // Ignore everything
  while (strcmp(Group.value,ENDSEC) != 0)
    In >> Group;
}


/* ------------------------------------------------------------------------ */


void ReadVertex() { // PolyLine vertices only

  int i;

  Vertex.Colour   = -1;
  Vertex.Type     = 0;
  Vertex.VCount   = 0;
  Vertex.Location = ZeroVector;
  for(i=0;i<4;i++) Vertex.Face[i] = 0;


  In >> Group;
  while (Group.code != 0) {
    switch (Group.code) {
      READ_ENTITY_OPTIONAL(Vertex);
      case 70: Assign(Vertex.Type,Group); break;
      case 71: Assign(Vertex.Face[0],Group); Vertex.VCount++; break;
      case 72: Assign(Vertex.Face[1],Group); Vertex.VCount++; break;
      case 73: Assign(Vertex.Face[2],Group); Vertex.VCount++; break;
      case 74: Assign(Vertex.Face[3],Group); Vertex.VCount++; break;
      READ_COORDINATE(Vertex.Location,10,20,30);
    }
    In >> Group;
  }
  Group.lastread = TRUE;
}


void ReadCircle() {

  Circle.Colour = -1;
  Circle.Thickness = 0;
  Circle.Normal.x = 0.0;
  Circle.Normal.y = 0.0;
  Circle.Normal.z = 1.0;

  In >> Group;
  while (Group.code != 0) {
    switch (Group.code) {
      READ_ENTITY_OPTIONAL(Circle);
      case 39: Assign(Circle.Thickness,Group); break;
      case 40: Assign(Circle.Radius,Group); break;
      READ_COORDINATE(Circle.Center,10,20,30);
      READ_ENTITY_NORMAL(Circle.Normal);
    }
    In >> Group;
  }
  Group.lastread = TRUE;
}


void Read3DFace() {

  Face3D.Colour = -1;

  In >> Group;
  while (Group.code != 0) {
    switch (Group.code) {
      READ_ENTITY_OPTIONAL(Face3D);
      READ_COORDINATE(Face3D.p[0],10,20,30);
      READ_COORDINATE(Face3D.p[1],11,21,31);
      READ_COORDINATE(Face3D.p[2],12,22,32);
      READ_COORDINATE(Face3D.p[3],13,23,33);
    }
    In >> Group;
  }
  Group.lastread = TRUE;
}

// Read in a polyline. If the polyline represents a 3D shape then read in
// the following vertices.

void ReadPolyLine() {
  int i,j,k,l,VerticesFollow;
  int M,N,M_wrap,N_wrap,Face_Count=0,Vertex_Count=0,fc;
  point p,p1,p2,p3,nold;

  PolyLine.Colour = -1;
  PolyLine.Type = 0;
  PolyLine.M_Count = 0;
  PolyLine.N_Count = 0;
  PolyLine.V_Count = 0;
  PolyLine.F_Count = 0;
  PolyLine.min     = MaxVector;
  PolyLine.max     = MinVector;

  In >> Group;
  while (Group.code != 0) {
    switch (Group.code) {
      READ_ENTITY_OPTIONAL(PolyLine);
      case 66: Assign(VerticesFollow,Group); break;
      case 70: Assign(PolyLine.Type,Group); break;
      case 71: Assign(PolyLine.M_Count,Group); break;
      case 72: Assign(PolyLine.N_Count,Group); break;
    }
    In >> Group;
  }

  N=PolyLine.N_Count,M=PolyLine.M_Count; Mesh[0] = ZeroVector;
  if (PolyLine.Type & 1)  M_wrap = 1; else M_wrap = 0;
  if (PolyLine.Type & 32) N_wrap = 1; else N_wrap = 0;

  // Read Vertices and faces, and calculate normals

  if (VerticesFollow) {

    Vertex_Count++;  // Vertices are indexed from 1
    while (strcmp(Group.value,SEQEND) != 0) {
      ReadVertex(); Group.lastread = FALSE;

      // Polyface mesh vertex OR Polygon mesh vertex
      if ((Vertex.Type & 128 && Vertex.Type & 64) || (Vertex.Type & 64)) {
	if (Vertex_Count < MAXMESH) {
	  Mesh[Vertex_Count]    = Vertex.Location;
	  Normals[Vertex_Count] = ZeroVector;

	  // Calculate extents
	  if (Mesh[Vertex_Count].x < PolyLine.min.x) PolyLine.min.x = Mesh[Vertex_Count].x;
	  if (Mesh[Vertex_Count].x > PolyLine.max.x) PolyLine.max.x = Mesh[Vertex_Count].x;
	  if (Mesh[Vertex_Count].y < PolyLine.min.y) PolyLine.min.y = Mesh[Vertex_Count].y;
	  if (Mesh[Vertex_Count].y > PolyLine.max.y) PolyLine.max.y = Mesh[Vertex_Count].y;
	  if (Mesh[Vertex_Count].z < PolyLine.min.z) PolyLine.min.z = Mesh[Vertex_Count].z;
	  if (Mesh[Vertex_Count].z > PolyLine.max.z) PolyLine.max.z = Mesh[Vertex_Count].z;
	}
	Vertex_Count++;
      }

      // Polyface mesh face
      if (Vertex.Type & 128 && !(Vertex.Type & 64)) {

	for (i=0;i<Vertex.VCount;i++) {
	  Faces[Face_Count][i] = Vertex.Face[i];
	  VCount[Face_Count]   = Vertex.VCount;
	}

	if (SmoothTriangles) { // Calculate normal
	  for (i=0;i<Vertex.VCount;i++) {
	    j = i-1; if (j==-1) j=Vertex.VCount-1;
	    nold = Normals[abs(Faces[Face_Count][i])];
	    p    = Mesh[abs(Faces[Face_Count][i])];
	    p1   = Mesh[abs(Faces[Face_Count][j])];
	    p2   = Mesh[abs(Faces[Face_Count][(i+1)%Vertex.VCount])];
	    p3   = Unit((p1-p)*(p-p2));
	    Normals[abs(Faces[Face_Count][i])] = Unit(p3+nold);
	  }
	}
	Face_Count++;
      }
    }
  }

  // Calculate normals for polygon mesh
  if (SmoothTriangles && PolyLine.Type && 16) {
    for (i=1;i<N+N_wrap;i++) {
      for (j=1;j<M+M_wrap;j++) {
	Fc[0] = (j%M)*N+(i%N)+1;
	Fc[1] = (j-1)%M*N+(i%N)+1;
	Fc[2] = (j-1)%M*N+(i-1)%N+1;
	Fc[3] = (j%M)*N+(i-1)%N+1;

	// Account for meshes that don't wrap
	if (!((!N_wrap && i==N) || (!M_wrap && j==M))) {

	  p    = Mesh[Fc[1]];
	  p1   = Mesh[Fc[0]];
	  p2   = Mesh[Fc[2]];
	  // Avoid degeneracy
	  if (p==p1) p1 = Mesh[Fc[3]];
	  else if (p==p2) p2 = Mesh[Fc[3]];
	  p3 = Unit((p2-p)*(p-p1));

	  // Assume all quads are planar since we
	  // don't know how they are triangulated.
	  for (k=0;k<4;k++) {
	    nold = Normals[Fc[k]];
	    Normals[Fc[k]] = Unit(p3+nold);
	  }
	}
      }
    }
  }

  PolyLine.V_Count = Vertex_Count;
  PolyLine.F_Count = Face_Count;

  if (GraphicsOn) {
    if (PolyLine.Type & 16)
      DrawPolygonMesh(Vertex_Count,M,N,M_wrap,N_wrap);
    else if (PolyLine.Type & 64)
      DrawPolyfaceMesh(Vertex_Count,Face_Count);

  setcolor(WHITE); outtextxy(0,470,"Press any key");
  char ch = getch();
  }
}

void ReadInsert() {

  point Ax,Ay,Rx,Ry,t;

  Insert.Attributes = 0;
  Insert.Rotation = 0;
  Insert.ColumnCount = 1;
  Insert.RowCount = 1;
  Insert.ColumnSpacing = 0;
  Insert.RowSpacing = 0;
  Insert.Scale.x = 1.0;
  Insert.Scale.y = 1.0;
  Insert.Scale.z = 1.0;
  Insert.Normal  = ZeroVector;

  In >> Group;
  while (Group.code != 0) {
    switch (Group.code) {
      READ_ENTITY_OPTIONAL(Insert);
      case 66: Assign(Insert.Attributes,Group); break;
      case 2:  Assign(Insert.Name,Group); break;
      case 50: Assign(Insert.Rotation,Group); break;
      case 70: Assign(Insert.ColumnCount,Group); break;
      case 71: Assign(Insert.RowCount,Group); break;
      case 44: Assign(Insert.ColumnSpacing,Group); break;
      case 45: Assign(Insert.RowSpacing,Group); break;
      READ_COORDINATE(Insert.Insertion,10,20,30);
      READ_COORDINATE(Insert.Scale,41,42,43);
      READ_COORDINATE(Insert.Normal,210,220,230);
    }
    In >> Group;
  }

  // Attribute entities

  if (Insert.Attributes == 1) {
    while (strcmp(Group.value,SEQEND) != 0)
      In >> Group;

    In >> Group;       			  // 8 <Layer>
    In >> Group;       			  // 5 <Handle>
  }

  if (Group.code == 0) Group.lastread = TRUE;

  // Evaluate entity co-ordinate system

  if (!(Insert.Normal == ZeroVector)) {
    if (fabs(Insert.Normal.x) < 1.0/64.0 && fabs(Insert.Normal.y) < 1.0/64.0)
      Ax = YUnit*Insert.Normal;
    else
      Ax = XUnit*Insert.Normal;
    Ay = Insert.Normal*Ax;

    Rx.x = cos(Insert.Rotation*PI/2.0);
    Rx.y = sin(Insert.Rotation*PI/2.0);
    Rx.z = 0;

    Ry = ZUnit*Rx;
  }

}


void ReadBlocks() {
  int BlocksEnd = FALSE;

  do {

    In >> Group;       			  // 0 BLOCK or 0 ENDSEC
    if (strcmp(Group.value,ENDSEC) == 0) BlocksEnd = TRUE;
    else {

      In >> Group;
      while (Group.code != 0) {
	switch (Group.code) {
	  READ_ENTITY_OPTIONAL(Block);
	  case 2:  Assign(Block.Name,Group); break;
	  case 70: Assign(Block.Type,Group); break;
	  READ_COORDINATE(Block.Base,41,42,43);
	}
	In >> Group;
      }

      if (!InExcludeList(Block.Name)) {
	WriteBlockStart(Block);

	Group.lastread = TRUE; ReadEntities(ENDBLK,Multiple);

	WriteBlockEnd(Block);
      }
      else {
	while(strcmp(Group.value,ENDBLK) != 0)
	  In >> Group;
      }

      In >> Group;     			  // 8 <Layer>
      In >> Group;     			  // 5 <Handle>
    }
  } while (BlocksEnd == FALSE);
}

void BlocksSection() {
  ReadBlocks();
}


/* ------------------------------------------------------------------------ */


void ReadEntities(char* Terminate,WriteMode Mode) {
  int EntitiesEnd = FALSE;

  do {
    In >> Group;     			  // 0 <Entity>

    if (strcmp(Group.value,Terminate) == 0) EntitiesEnd = TRUE;
    else if (strcmp(Group.value,CIRCLE) == 0)
      { ReadCircle(); WriteCircleEntity(Circle,Mode); }
    else if (strcmp(Group.value,FACE3D) == 0)
      { Read3DFace(); WriteTriangleEntity(Face3D,Mode); }
    else if (strcmp(Group.value,POLYLINE) == 0)
      { ReadPolyLine(); WriteMesh(PolyLine,Mesh,Normals,Faces,VCount,Mode); }
    else if (strcmp(Group.value,INSERT) == 0)
      { ReadInsert(); if (!InExcludeList(Insert.Name)) WriteInsertEntity(Insert); }
  } while (EntitiesEnd == FALSE);
}

void EntitiesSection() {
  ReadEntities(ENDSEC,Single);
}


/* ------------------------------------------------------------------------ */


void main (int argc, char *argv[]) {

  int i;

  cout << "\nDXF to POV converter (Version 0.81) : March-April 1994\n";
  cout << "Created by Richard Dorman\n";

  // Check number of arguments.
  if (argc < 3) {
    cerr << "\nSyntax : DXF-POV <input file> <output file> [+/-options]\n\n";
    cerr << "where [options] can be (defaults in []) : \n";
    cerr << "\tcamera   [+] : camera on/off\n";
    cerr << "\tlayer    [+] : layer name conversion on/off\n";
    cerr << "\tsmooth   [+] : mesh triangle smoothing\n";
    cerr << "\tgraphics [-] : graphics on/off\n";
    cerr << "\nExample : DXF-POV spanner spanner -c +l\n\n";
    exit(1);
  }

  // Get file names and add extensions if necessary. Convert to upper case.
  stpcpy(Inputfile,argv[1]);
  if (strcspn(Inputfile,".") == strlen(Inputfile))
    strcat(Inputfile,".DXF");
  stpcpy(Outputfile,argv[2]);
  if (strcspn(Outputfile,".") == strlen(Outputfile))
    strcat(Outputfile,".INC");
  strupr(Inputfile); strupr(Outputfile);

  // Handle command line options.
  for (i=3;i<argc;i++) {
    strlwr(argv[i]);
    if (argv[i][0] == '-') {
      if (argv[i][1] == CAMERA_OPTION)   CameraOn = FALSE; else
      if (argv[i][1] == CONVERT_OPTION)  ConvertLayerNames = FALSE; else
      if (argv[i][1] == SMOOTH_OPTION)   SmoothTriangles = FALSE; else
      if (argv[i][1] == GRAPHICS_OPTION) GraphicsOn = FALSE;
    }
    else if (argv[i][0] == '+') {
      if (argv[i][1] == CAMERA_OPTION)   CameraOn = TRUE; else
      if (argv[i][1] == CONVERT_OPTION)  ConvertLayerNames = TRUE; else
      if (argv[i][1] == SMOOTH_OPTION)   SmoothTriangles = TRUE; else
      if (argv[i][1] == GRAPHICS_OPTION) GraphicsOn = TRUE;
    }
    else
      cerr << "Unknown option : " << argv[i] << "\n";
  }

  // Open files and check for errors.
  In.open(Inputfile);
  if (!In) { cerr << "\nCannot open " << Inputfile << " for input\n"; exit(1); }
  Out.open(Outputfile);
  if (!Out) { cerr << "\nCannot open " << Outputfile << " for output\n"; exit(1); }

  if (GraphicsOn) {
    registerbgidriver(EGAVGA_driver);
    int gdriver=VGA,gmode=VGAHI;
    initgraph(&gdriver,&gmode,"");
  }

  // Initialise group
  Group.line = 0;
  Group.lastread = FALSE;

  // Read DXF file and write POV file for each section.

  In >> Group;
  while (strcmp(Group.value,FILEEND) != 0) {

    In >> Group;
    if (strcmp(Group.value,HEADER) == 0)   HeaderSection();
    if (strcmp(Group.value,TABLES) == 0)   TablesSection();
    if (strcmp(Group.value,BLOCKS) == 0)   BlocksSection();
    if (strcmp(Group.value,ENTITIES) == 0) EntitiesSection();
    In >> Group;

  }

  if (GraphicsOn) closegraph();

  Out.close();
  In.close();
}