// DXF to POV converter (version 0.81) : Header file
// Coded by Richard Dorman (March-April 1994)

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define FALSE 0
#define TRUE  1
#define MAXSTRING 256
#define MAXMESH   350
#define MAXFLOAT  3.4E38
#define MINFLOAT  -3.4E-38
#define PI        3.1415927

#define Magnitude(v) sqrt(v.x*v.x+v.y*v.y+v.z*v.z)   // |v|
#define Dot(v1,v2)   (v1.x*v2.x+v1.y*v2.y+v1.z*v2.z) // v1.v2

extern ifstream In;
extern ofstream Out;
extern CameraOn;
extern ConvertLayerNames;
extern SmoothTriangles;
extern GraphicsOn;

enum WriteMode {Single,Multiple};
enum EndMode   {Open,Closed};

typedef struct {
  int   lastread;
  int   line;
  int   code;
  char  value[MAXSTRING];
} Group_Type;

typedef struct { float x,y,z; } point;

typedef struct {
  char  Name[MAXSTRING];
  point Shift;
  point LookAt;
  point Location;
  float TwistAngle;
  float Lens;
} ViewPort_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  int   Type;
  point Location;
  int   Face[4];  // Four vertices of a face. No invisible edges.
  int   VCount;   // How many vertices in the face.
} Vertex_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  point Center;
  float Radius;
  point Normal;
  float Thickness;
} Circle_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  point p[4];     // No invisible edges.
} Face3D_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  int   Type;
  int   M_Count;
  int   N_Count;
  int   V_Count;  // Vertex count
  int   F_Count;  // Face count
  point min,max;  // Extents
} PolyLine_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  int   Attributes;
  char  Name[MAXSTRING];
  point Insertion;
  point Scale;
  point Normal;
  float Rotation;
  int   ColumnCount;
  int   RowCount;
  float ColumnSpacing;
  float RowSpacing;
} Insert_Type;

typedef struct {
  char  Layer[MAXSTRING];
  char  Handle[MAXSTRING];
  int   Colour;
  char  Name[MAXSTRING];
  int   Type;
  point Base;
  // No XRef
} Block_Type;

istream& operator >>(istream&,Group_Type&);
void Assign(char*,Group_Type&);
void Assign(float&,Group_Type&);
void Assign(int&,Group_Type&);

extern point ZeroVector;
extern point XUnit;
extern point YUnit;
extern point ZUnit;
extern point MinVector;
extern point MaxVector;
point Unit(point);
int   operator ==(point,point);
point operator +(point,point);
point operator -(point,point);
point operator *(point,point);
point operator *(float,point);

void HeaderSection();
void TablesSection();
void BlocksSection();
void EntitiesSection();

void ReadBlocks();
void ReadVertex();
void ReadCircle();
void Read3DFace();
void ReadPolyLine();
void ReadInsert();
void ReadEntities(char*,WriteMode);


void WriteBlockStart(Block_Type);
void WriteCamera(ViewPort_Type);
void WriteTriangle(Face3D_Type,int);
void WriteSmoothTriangle(Face3D_Type,Face3D_Type,int);
void WriteTexture(char*);
void WriteInsertEntity(Insert_Type);
void WriteCircleEntity(Circle_Type,WriteMode);
void WriteTriangleEntity(Face3D_Type,WriteMode);
void WriteMesh(PolyLine_Type,point[MAXMESH],point[MAXMESH],int[MAXMESH][4],int[MAXMESH],WriteMode);
void WriteBlockEnd(Block_Type);
