// DXF to POV converter (version 0.81) : POV output module
// Coded by Richard Dorman (March-April 1994)

// Notes : Scales are always taken as absolute values. Negative values
//         cause strange results.

#include "DXF-POV.h"

#define DECLARE        "#declare "
#define UNION          "union {\n"
#define TEXTURE        "texture { "
#define END            "}\n"
#define TRIANGLE       "triangle { "
#define SMOOTHTRIANGLE "smooth_triangle { "
#define DISC           "disc { "
#define CYLINDER       "cylinder { "
#define OBJECT         "object {\n"
#define TRANSLATE      "translate "
#define SCALE          "scale "
#define CAMERA         "camera {\n"
#define BOUNDEDBYBOX   "bounded_by { box { "
#define LOCATION       "location "
#define SKY            "sky "
#define RIGHT          "right "
#define LOOKAT         "look_at "

#define DEFAULT_TEXTURE "pigment { color red 0.5 green 0.5 blue 0.5 }"
#define LAYER0    "0"

#define COORDINATE(A) "<" << A.x << "," << A.y << "," << A.z << ">"

char *indent[3] = {"","  ","    "};
int index = 0;

void WriteBlockStart(Block_Type Block) {
  Out << DECLARE << Block.Name << " = " << UNION;
  index++;
}

void WriteCamera(ViewPort_Type View) {
  if (CameraOn) {
    Out << "// Camera : " << View.Name << "\n";
    Out << indent[index++] << CAMERA;
    Out << indent[index] << LOCATION << "<"
	<< View.Location.x+View.Shift.x << ","
	<< View.Location.y+View.Shift.y << ","
	<< View.Location.z << ">" << "\n";
    // Out << indent[index] << SKY << "<0,0,1>\n";
    // Out << indent[index] << RIGHT << "<-1.3333,0,0>\n";
    Out << indent[index] << LOOKAT   << "<"
	<< View.LookAt.x+View.Shift.x << ","
	<< View.LookAt.y+View.Shift.y << ","
	<< View.LookAt.z << ">" << "\n";
    Out << indent[--index] << END << "\n";
  }
}

void WriteBoundedBy(point Min,point Max) {
  Out << indent[index] << BOUNDEDBYBOX << COORDINATE(Min)
				       << "," COORDINATE(Max) << " } " << END;
}

void WriteTriangle(Face3D_Type Face,int Vertices) {

  if (Vertices < 3 )
    cerr << "Warning : Too few vertices in Smooth Polgon. Ignored.\n";
  else if (Vertices > 4)
    cerr << "Warning : Too many vertices in Smooth Polygon. Ignored.\n";
  else {
    // Check for degeneracy
    if (!(Face.p[0] == Face.p[1]) && !(Face.p[1] == Face.p[2]) && !(Face.p[2] == Face.p[0])) {
      Out << indent[index] << TRIANGLE;
      Out << COORDINATE(Face.p[0]) << ", ";
      Out << COORDINATE(Face.p[1]) << ", ";
      Out << COORDINATE(Face.p[2]) << END;
    }
    if (!(Face.p[0] == Face.p[2]) && !(Face.p[2] == Face.p[3]) && !(Face.p[3] == Face.p[0]) &&
      Vertices == 4) {
      Out << indent[index] << TRIANGLE;
      Out << COORDINATE(Face.p[0]) << ", ";
      Out << COORDINATE(Face.p[2]) << ", ";
      Out << COORDINATE(Face.p[3]) << END;
    }
  }
}

void WriteSmoothTriangle(Face3D_Type Face,Face3D_Type Normal,int Vertices) {

  if (Vertices < 3 )
    cerr << "Warning : Too few vertices in Smooth Polgon. Ignored.\n";
  else if (Vertices > 4)
    cerr << "Warning : Too many vertices in Smooth Polygon. Ignored.\n";
  else {
    // Check for degeneracy
    if (!(Face.p[0] == Face.p[1]) && !(Face.p[1] == Face.p[2]) && !(Face.p[2] == Face.p[0])) {
      Out << indent[index] << SMOOTHTRIANGLE;
      Out << COORDINATE(Face.p[0]) << ", " << COORDINATE(Normal.p[0]) << ", ";
      Out << COORDINATE(Face.p[1]) << ", " << COORDINATE(Normal.p[1]) << ", ";
      Out << COORDINATE(Face.p[2]) << ", " << COORDINATE(Normal.p[2]) << END;
    }
    if (!(Face.p[0] == Face.p[2]) && !(Face.p[2] == Face.p[3]) && !(Face.p[3] == Face.p[0]) &&
	Vertices == 4) {
      Out << indent[index] << SMOOTHTRIANGLE;
      Out << COORDINATE(Face.p[0]) << ", " << COORDINATE(Normal.p[0]) << ", ";
      Out << COORDINATE(Face.p[2]) << ", " << COORDINATE(Normal.p[2]) << ", ";
      Out << COORDINATE(Face.p[3]) << ", " << COORDINATE(Normal.p[3]) << END;
    }
  }
}


void TextureConvert(char Texture[256]) {
  int pos=0,capital=TRUE;

  while (Texture[pos] != '\0') {
    if (capital)
      capital = FALSE;
    else if (Texture[pos] >= 'A' && Texture[pos] <= 'Z')
      Texture[pos] = Texture[pos]+32;
    else
      capital = TRUE;
    pos++;
  }
}

void WriteTexture(char* Layer) {
  double value=MINFLOAT; char *endptr; int test;

  value = strtod(Layer, &endptr);
  test  = strcmp(Layer,LAYER0);
  if (ConvertLayerNames) TextureConvert(Layer);
  if (value == 0.0 && test != 0)
    Out << indent[index] << TEXTURE << Layer << " " << END;
  else
    Out << indent[index] << TEXTURE << DEFAULT_TEXTURE << " " << END;
}

void WriteInsertEntity(Insert_Type Insert) {
  Out << indent[index++] << OBJECT;
  Out << indent[index] << Insert.Name << "\n";
  if (Insert.Scale.x != 1 || Insert.Scale.y != 1 || Insert.Scale.z != 1)
    Out << indent[index] << SCALE << "<" << fabs(Insert.Scale.x) << ","
					 << fabs(Insert.Scale.y) << ","
					 << fabs(Insert.Scale.z) << ">" << "\n";
  if (Insert.Insertion.x != 0 || Insert.Insertion.y != 0 || Insert.Insertion.z != 0)
    Out << indent[index] << TRANSLATE << COORDINATE(Insert.Insertion) << "\n";

  // No rotate yet

  WriteTexture(Insert.Layer);
  Out << indent[--index] << END << "\n";
}

void WriteCylinder(Circle_Type Circle,EndMode End) {
  point p;

  // First calculate other end for cylinder
  p.x = Circle.Normal.x*Circle.Thickness+Circle.Center.x;
  p.y = Circle.Normal.y*Circle.Thickness+Circle.Center.y;
  p.z = Circle.Normal.z*Circle.Thickness+Circle.Center.z;

  Out << indent[index] << CYLINDER << COORDINATE(Circle.Center) << ", "
		       << COORDINATE(p) << ", " << Circle.Radius;
  if (End == Open) Out << " open";
  Out << indent[index] << END;
}


void WriteCircleEntity(Circle_Type Circle,WriteMode Mode) {
  if (Mode == Single) Out << indent[index++] << OBJECT;
  if (Circle.Thickness == 0.0)
    Out << indent[index] << DISC << COORDINATE(Circle.Center) << ", "
			 << COORDINATE(Circle.Normal) << ", "
			 << Circle.Radius << " " << END;
  else
    WriteCylinder(Circle,Closed);
  if (Mode == Single) {
    WriteTexture(Circle.Layer);
    Out << indent[--index] << END << "\n";
  }
}

void WriteTriangleEntity(Face3D_Type Face3D,WriteMode Mode) {
  if (Mode == Single) Out << indent[index++] << OBJECT;
  WriteTriangle(Face3D,4);
  if (Mode == Single) {
    WriteTexture(Face3D.Layer);
    Out << indent[--index] << END << "\n";
  }
}

void WriteMesh(PolyLine_Type PolyLine,point Mesh[MAXMESH],point Normals[MAXMESH],
	       int Faces[MAXMESH][4],int VCount[MAXMESH],WriteMode Mode) {

  int i,j,N,M,N_wrap,M_wrap;
  Face3D_Type Face,Normal;

  if (PolyLine.Type & 16 || PolyLine.Type & 64) {
    if (Mode == Single) {
      Out << indent[index++] << OBJECT;
      Out << indent[index++] << UNION;
    }

    N=PolyLine.N_Count,M=PolyLine.M_Count;
    if (PolyLine.Type & 1)  M_wrap = 1; else M_wrap = 0;
    if (PolyLine.Type & 32) N_wrap = 1; else N_wrap = 0;

    // Polyface mesh : regular mesh
    if (PolyLine.Type & 16) {
      for (i=1;i<N+N_wrap;i++)
	for (j=1;j<M+M_wrap;j++) {
	  Face.p[0] = Mesh[(j%M)*N+(i%N)+1];     Normal.p[0] = Normals[(j%M)*N+(i%N)+1];
	  Face.p[1] = Mesh[(j-1)%M*N+(i%N)+1];   Normal.p[1] = Normals[(j-1)%M*N+(i%N)+1];
	  Face.p[2] = Mesh[(j-1)%M*N+(i-1)%N+1]; Normal.p[2] = Normals[(j-1)%M*N+(i-1)%N+1];
	  Face.p[3] = Mesh[(j%M)*N+(i-1)%N+1];   Normal.p[3] = Normals[(j%M)*N+(i-1)%N+1];
	  if (SmoothTriangles)
	    WriteSmoothTriangle(Face,Normal,4);
	  else
	    WriteTriangle(Face,4);
	}
     }
     // Polygon mesh : irregualr mesh
     else if (PolyLine.Type & 64) {
       for (i=0;i<PolyLine.F_Count;i++) {
	 for (j=0;j<VCount[i];j++) {
	   Face.p[j] = Mesh[abs(Faces[i][j])];
	   Normal.p[j] = Normals[abs(Faces[i][j])];
	 }
	 if (SmoothTriangles)
	   WriteSmoothTriangle(Face,Normal,VCount[i]);
	 else
	    WriteTriangle(Face,VCount[i]);
       }
     }

    if (Mode == Single) {
      WriteTexture(PolyLine.Layer);
      Out << indent[--index] << END;         // Union
      Out << indent[--index] << END << "\n"; // Object
    }
  }
}

void WriteBlockEnd(Block_Type Block) {
  WriteTexture(Block.Layer);
  Out << END << "\n";
  index--;
}
